
import sqlalchemy
from grokcore import component as grok

from zope.component import provideAdapter
from zope.interface import Interface
from zeam.form.base.interfaces import IFieldFactory
from zeam.form.ztk import widgets as fields
from zeam.form.rdb.source import foreign_source

class IFieldBuilder(Interface):
    pass


class SQLFieldBuilder(object):
    grok.implements(IFieldBuilder)
    field = fields.TextLineField
    sql_field = sqlalchemy.types.String

    def __init__(self, sql_type):
        self.sql_type = sql_type

    @classmethod
    def register(cls):
        provideAdapter(cls, (cls.sql_field,), IFieldBuilder)

    def __call__(self, sql_field):
        options = sql_field.info.copy()
        instance_options = {}
        if 'identifier' not in options:
            options['identifier'] = sql_field.name
        if 'title' not in options:
            options['title'] = unicode(sql_field.description)
        if 'required' not in options:
            options['required'] = not sql_field.nullable
        if 'defaultValue' not in options:
            if sql_field.default is not None:
                options['defaultValue'] = sql_field.default.arg
        if 'mode' in options:
            instance_options['mode'] = options['mode']

        if len(sql_field.foreign_keys):
            foreign_key = sql_field.foreign_keys[0]
            factory = fields.ChoiceField
            options['source'] = foreign_source(foreign_key.column, options)
        elif 'factory' in options:
            factory = options['factory']
            del options['factory']
        else:
            factory = self.field
        field = factory(**options)
        for name, value in instance_options.iteritems():
            setattr(field, name, value)
        return field


class StringFieldBuilder(SQLFieldBuilder):
    pass


class TextFieldBuilder(SQLFieldBuilder):
    field = fields.TextField
    sql_field = sqlalchemy.types.Text


class IntegerFieldBuilder(SQLFieldBuilder):
    field = fields.IntegerField
    sql_field = sqlalchemy.types.Integer


class FloatFieldBuilder(SQLFieldBuilder):
    field = fields.FloatField
    sql_field = sqlalchemy.types.Float


class DateFieldBuilder(SQLFieldBuilder):
    field = fields.DateField
    sql_field = sqlalchemy.types.Date


class DatetimeFieldBuilder(SQLFieldBuilder):
    field = fields.DatetimeField
    sql_field = sqlalchemy.types.DateTime


class TimeFieldBuilder(SQLFieldBuilder):
    field = fields.TimeField
    sql_field = sqlalchemy.types.Time


class BooleanFieldBuilder(SQLFieldBuilder):
    field = fields.BooleanField
    sql_field = sqlalchemy.types.Boolean


class ColumnFieldFactory(object):
    """Generate zeam.form fields out of a SQLAlchemy column.
    """
    grok.implements(IFieldFactory)

    def __init__(self, column):
        self.column = column

    def produce(self):
        factory = IFieldBuilder(self.column.type)
        yield factory(self.column)


class TableFieldFactory(object):
    """Generate zeam.form fields out of a SQLAlchemy table.
    """
    grok.implements(IFieldFactory)

    def __init__(self, table):
        self.table = table

    def produce(self):
        for column in self.table.columns:
            if column.primary_key:
                continue
            factory = IFieldBuilder(column.type)
            yield factory(column)


def register():
    # Factories
    StringFieldBuilder.register()
    TextFieldBuilder.register()
    IntegerFieldBuilder.register()
    FloatFieldBuilder.register()
    DateFieldBuilder.register()
    DatetimeFieldBuilder.register()
    TimeFieldBuilder.register()
    BooleanFieldBuilder.register()
    # Field producer
    provideAdapter(
        ColumnFieldFactory,
        (sqlalchemy.Column,))
    provideAdapter(
        TableFieldFactory,
        (sqlalchemy.Table,))




from z3c.saconfig import Session
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.i18nmessageid import MessageFactory


_ = MessageFactory('zeam.form.base')


class ExtendingVocabulary(SimpleVocabulary):
    """A vocabulary that can be extended with terms of an another
    vocabulary on demand.
    """

    def __init__(self, terms, query=None):
        super(ExtendingVocabulary, self).__init__(terms)
        self._query = query
        self._vocabulary = None

    def addTerm(self, term):
        if term.value in self.by_value:
            raise ValueError(term)
        if term.token in self.by_token:
            raise ValueError(term)
        self.by_value[term.value] = term
        self.by_token[term.token] = term
        self._terms.append(term)

    def getVocabulary(self):
        if self._vocabulary is None and self._query is not None:
            self._vocabulary = SimpleVocabulary(list(self._query))
        return self._vocabulary

    def getTerm(self, value):
        try:
            return super(ExtendingVocabulary, self).getTerm(value)
        except LookupError:
            vocabulary = self.getVocabulary()
            if vocabulary is None:
                raise
            term = vocabulary.getTerm(value)
            self.addTerm(term)
            return term

    def getTermByToken(self, token):
        try:
            return super(ExtendingVocabulary, self).getTermByToken(token)
        except LookupError:
            vocabulary = self.getVocabulary()
            if vocabulary is None:
                raise
            term = vocabulary.getTermByToken(token)
            self.addTerm(term)
            return term


def foreign_source(value_column, options):
    """Generate a vocabulary containing terms for each possible value
    you can set in a SQLAlchemy foreign key.
    """
    defaults = []
    if not options['required']:
        defaults.append(SimpleTerm(title=_(u'(not set)'), value=None))
    session = Session()

    table = value_column.table
    columns = [value_column,]
    if 'foreignTitleColumn' in options:
        if isinstance(options['foreignTitleColumn'], basestring):
            columns.append(table.columns[options['foreignTitleColumn']])
        else:
            for identifier in options['foreignTitleColumn']:
                columns.append(table.columns[identifier])
        del options['foreignTitleColumn']
    else:
        columns.append(value_column)
    if 'foreignTitleFactory' in options:
        title_factory = options['foreignTitleFactory']
        del options['foreignTitleFactory']
    else:
        title_factory = lambda d: str(d[0])

    def create_terms(query):
        for result in query.all():
            yield SimpleTerm(
                title=title_factory(result[1:]),
                value=result[0])

    full_query = session.query(*columns)
    if 'foreignTitleQuery' in options:
        title_terms = create_terms(options['foreignTitleQuery'](full_query))
        full_terms = create_terms(full_query)
        del options['foreignTitleQuery']
    else:
        title_terms = create_terms(full_query)
        full_terms = None

    return ExtendingVocabulary(defaults + list(title_terms), full_terms)





from zeam.form.base import Fields

class ModelFields(object):
    """Object descriptor that generate and cache a list of zeam.form
    fields out of a SQLAlchemy model that it is applied on.
    """

    def __new__(cls, model=None):
        if model is not None:
            return Fields(model.__table__)
        return object.__new__(cls, model=model)

    def __init__(self, model=None):
        self.__cache = None

    def __get__(self, form, type=None):
        if form is not None:
            model = form.context.__class__
            return Fields(model.__table__)
        return Fields()

    def __set__(self, form, value):
        raise AttributeError

